@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: LGPL-2.1-or-later
 /*
  *
  *  BlueZ - Bluetooth protocol stack for Linux
@@ -5,20 +6,6 @@
  *  Copyright (C) 2014  Intel Corporation. All rights reserved.
  *
  *
- *  This library is free software; you can redistribute it and/or
- *  modify it under the terms of the GNU Lesser General Public
- *  License as published by the Free Software Foundation; either
- *  version 2.1 of the License, or (at your option) any later version.
- *
- *  This library is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- *  Lesser General Public License for more details.
- *
- *  You should have received a copy of the GNU Lesser General Public
- *  License along with this library; if not, write to the Free Software
- *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
- *
  */
 
 #ifdef HAVE_CONFIG_H
@@ -177,14 +164,6 @@ static struct device *device_create(const bdaddr_t *bdaddr)
 	struct device *dev;
 
 	dev = new0(struct device, 1);
-	if (!dev)
-		return NULL;
-
-	if (!queue_push_tail(devices, dev)) {
-		error("hf-client: Could not push dev on the list");
-		free(dev);
-		return NULL;
-	}
 
 	bacpy(&dev->bdaddr, bdaddr);
 	dev->state = HAL_HF_CLIENT_CONN_STATE_DISCONNECTED;
@@ -192,6 +171,8 @@ static struct device *device_create(const bdaddr_t *bdaddr)
 
 	init_codecs(dev);
 
+	queue_push_tail(devices, dev);
+
 	return dev;
 }
 
@@ -2166,10 +2147,6 @@ bool bt_hf_client_register(struct ipc *ipc, const bdaddr_t *addr)
 	DBG("");
 
 	devices = queue_new();
-	if (!devices) {
-		error("hf-client: Could not create devices list");
-		goto failed;
-	}
 
 	bacpy(&adapter_addr, addr);
 
